# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_poisson_solvers.ipynb.

# %% auto 0
__all__ = ['poisson_gauss_seidel_with_sor_solver', 'poisson_non_iterative_solver', 'poisson_newton_solver',
           'poisson_newton_alternative_solver']

# %% ../nbs/01_poisson_solvers.ipynb 5
from . import core

from functools import partial

import numpy as np
import scipy
from scipy import sparse


# %% ../nbs/01_poisson_solvers.ipynb 9
def poisson_gauss_seidel_with_sor_solver(
    w, r=None, verbose=False, log_middle_values=False,
):
    N = w.shape[0] - 1
    nx = ny = N + 1

    h = 1 / N

    # Handle the SOR parameter
    if r is None:
        r = 2 / (1 + np.pi / N) # optimal value

    psi = np.zeros((nx, ny)) # streamfunction

    middle_values = []

    for iteration in range(1, 4 * nx + 1):
        for i in range(1, nx - 1):
            for j in range(1, ny - 1):
                psi[i, j] = (1 - r) * psi[i, j] + r * (
                    psi[i - 1, j] + psi[i + 1, j] +
                    psi[i, j - 1] + psi[i, j + 1] +
                    w[i, j] * h ** 2
                ) / 4
        
        if verbose:
            print(f"{iteration=}; psi(0.5, 0.5) = {psi[nx // 2, ny // 2]}")
        
        if log_middle_values:
            middle_values.append(psi[nx // 2, ny // 2])
    
    return psi, iteration, middle_values


# %% ../nbs/01_poisson_solvers.ipynb 24
def construct_laplacian_kernel_matrix(N, h=None):
    '''
    Construct the matrix that defines the linear system in sparse form
    '''

    if h is None:
        h = 1 / N
    
    # The output size is N - 1 because we are excluding the boundary points
    N -= 1

    values = np.ones(5 * N**2 - 4 * N) / h**2
    values[:N**2] *= -4

    all_indices = np.arange(N ** 2)
    
    rows = [np.arange(N ** 2)]
    cols = [np.arange(N ** 2)]

    rows.append(all_indices[N:N**2])
    cols.append(all_indices[:-N])

    rows.append(all_indices[:-N])
    cols.append(all_indices[N:N**2])

    rows.append(all_indices[all_indices % N != 0])
    cols.append(all_indices[(all_indices + 1) % N != 0])

    rows.append(all_indices[(all_indices + 1) % N != 0])
    cols.append(all_indices[all_indices % N != 0])

    # Flatten the lists
    rows = np.concatenate(rows)
    cols = np.concatenate(cols)

    # Create the sparse matrix from the above information
    return sparse.csr_matrix((values, (rows, cols)), shape=(N ** 2, N ** 2))


# %% ../nbs/01_poisson_solvers.ipynb 25
def poisson_non_iterative_solver(w, algorithm="bicgstab"):
    N = w.shape[0] - 1

    # Construct the matrix that defines the linear system
    kernel_matrix = construct_laplacian_kernel_matrix(N)
    
    # Cast vorticity to the required form
    w = w[1:-1, 1:-1].flatten()

    # Solve the linear system.
    # Note -ve sign is applied to make A positive definite which is required
    # by some of the iterative solvers.
    psi = core.solve_sparse_linear_system(
        A=-kernel_matrix, b=w, algorithm=algorithm
    )
    
    psi = psi.reshape(N-1, N-1) # exclude the boundary points

    psi = np.pad(psi, (1, 1), mode="constant")
    
    return psi


# %% ../nbs/01_poisson_solvers.ipynb 32
def get_jacobian(N, h=1):
    return construct_laplacian_kernel_matrix(N, h=h)
    

def f(x, w):
    N = int(np.sqrt(x.shape[0])) + 1
    h = 1 / N

    x = x.reshape(N-1, N-1)

    # note we compute h ** 2 * f_current
    f = -4 * x + h ** 2 * w[1:-1, 1:-1]
    f[1:, :] += x[:-1, :]
    f[:-1, :] += x[1:, :]
    f[:, :-1] += x[:, 1:]
    f[:, 1:] += x[:, :-1]

    return f.flatten()


def newton_iterator(
    w, f, get_jacobian,
    algorithm="bicgstab", TOL=1e-8, max_iter=10, quiet=True
):
    '''
        - w: vorticity
        - f: evaluates the function given x, w
        - get_jacobian: evaluates the Jacobian given N
    '''

    N = w.shape[0] - 1
    
    n_iter = 0 # number of iterations

    # Initialization
    x = np.zeros(((N - 1) ** 2))
    f_current = f(x, w)
    
    # Check if the initial guess is a solution
    f_norm = scipy.linalg.norm(f_current)
    if f_norm <= TOL:
        if not quiet:
            print(f"n_iter={n_iter}")

        return x, n_iter
    
    
    while n_iter < max_iter:
        n_iter += 1
        # Set h=1 to have better scaling and note that we actually compute h ** 2 * f_current
        jacobian = get_jacobian(N)
        
        dx = core.solve_sparse_linear_system(
            A=jacobian, b=-f_current, algorithm=algorithm
        )
        x_next = x + dx
        
        f_current = f(x_next, w)
        
        f_norm = scipy.linalg.norm(f_current)
        if not quiet:
            print(f"iter={n_iter}; |residual|={f_norm}; |dx|={scipy.linalg.norm(dx)}")
        if f_norm <= TOL:
            break
        
        x = x_next

    if not quiet:
        print(f"n_iter={n_iter}")
    
    return x_next, n_iter


# %% ../nbs/01_poisson_solvers.ipynb 33
def poisson_newton_solver(
    w, algorithm="bicgstab", TOL=1e-8, max_iter=10, quiet=True
):
    solution, _ = newton_iterator(
        w=w, f=f, get_jacobian=get_jacobian,
        algorithm=algorithm,
        TOL=TOL, max_iter=max_iter, quiet=quiet
    )
    
    N = w.shape[0] - 1
    psi = solution.reshape(N - 1, N - 1)

    psi = np.pad(psi, (1, 1), mode="constant", constant_values=0)
    
    return psi


# %% ../nbs/01_poisson_solvers.ipynb 34
def poisson_newton_alternative_solver(w, **kwargs):
    N = w.shape[0] - 1

    # Can define jacobian as `jac=lambda x, args: get_jacobian(N=N, h=1/N)`
    solution = scipy.optimize.root(
        fun=f, x0=np.zeros((N - 1) ** 2),
        args=(w, ),
        **kwargs
    )
    
    psi = solution.x.reshape(N - 1, N - 1)

    psi = np.pad(psi, (1, 1), mode="constant", constant_values=0)
    
    return psi, solution




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% auto 0
__all__ = ['setup_poisson_problem', 'solve_sparse_linear_system', 'solve_sparse_linear_system_pytorch',
           'solve_sparse_linear_system_jax', 'calculate_force', 'calculate_velocity']

# %% ../nbs/00_core.ipynb 4
from functools import partial

import numpy as np
from scipy import sparse

import torch

import jax
import lineax as lx


# %% ../nbs/00_core.ipynb 6
def setup_poisson_problem(N):
    """
    Setup a 2D Poisson problem. Only square domains are supported.
    """

    nx = ny = N + 1

    # Initialize grid
    x_grid, y_grid = np.meshgrid(
        np.linspace(0, 1, nx), np.linspace(0, 1, ny), indexing="ij"
    )

    # Initialize the value of the vorticity on the grid
    w = 2 * np.pi ** 2 * np.sin(np.pi * x_grid) * np.sin(np.pi * y_grid)

    exact_solution = np.sin(np.pi * x_grid) * np.sin(np.pi * y_grid)

    return w, exact_solution, nx, ny, x_grid, y_grid


# %% ../nbs/00_core.ipynb 7
SPARSE_ALGORITHM_DICT = {
    "base": sparse.linalg.spsolve,
    "cg": sparse.linalg.cg,
    "bicgstab": sparse.linalg.bicgstab,
}


# %% ../nbs/00_core.ipynb 8
def solve_sparse_linear_system(A, b, algorithm="base", **kwargs):
    """
    Solve a sparse linear system using the specified algorithm.
    """

    solver = SPARSE_ALGORITHM_DICT[algorithm]

    if algorithm == "base" and len(b.shape) == 1:
        b = sparse.csr_matrix(b[:, None])

    soln = solver(A, b, **kwargs)

    return soln if algorithm == "base" else soln[0]


# %% ../nbs/00_core.ipynb 9
def lu_solve(A, b):
    LU, pivots = torch.linalg.lu_factor(A)
    return torch.linalg.lu_solve(LU, pivots, b)


TORCH_ALGORITHM_DICT = {
    "base": torch.linalg.solve,
    "lu_solve": lu_solve,
    "lstsq": torch.linalg.lstsq,
}


# %% ../nbs/00_core.ipynb 10
def solve_sparse_linear_system_pytorch(A, b, algorithm="base", **kwargs):
    """
    Solve a sparse linear system using the specified algorithm.
    """

    solver = TORCH_ALGORITHM_DICT[algorithm]

    out = solver(A, b, **kwargs)

    if algorithm == "lstsq": out = out.solution

    return out


# %% ../nbs/00_core.ipynb 11
JAX_ALGORITHM_DICT = {
    "jax_base": jax.numpy.linalg.solve,
    # "jax_base_scipy": jax.scipy.sparse.linalg.solve, # NOTE: slower than jax_base
    # "jax_bicgstab": jax.scipy.sparse.linalg.bicgstab, # NOTE: unstable
    # "jax_gmres": jax.scipy.sparse.linalg.gmres, # NOTE: unstable
    # "jax_lstsq": jax.numpy.linalg.lstsq, # NOTE: too slow
    "lineax_lu": partial(lx.linear_solve, solver=lx.LU()), # NOTE: same as lineax.AutoLinearSolver
    # "lineax_qr": partial(lx.linear_solve, solver=lx.QR()), # NOTE: slower than jax_base
    # "lineax_svd": partial(lx.linear_solve, solver=lx.SVD()), # NOTE: too slow
    # "lineax_bicgstab": partial(
    #     lx.linear_solve, solver=lx.BiCGStab(rtol=1e-6, atol=1e-6)
    # ), # NOTE: unstable
    # "lineax_gmres": partial(
    #     lx.linear_solve, solver=lx.GMRES(rtol=1e-6, atol=1e-6)
    # ), # NOTE: unstable
}


# %% ../nbs/00_core.ipynb 12
def solve_sparse_linear_system_jax(A, b, algorithm="jax_base", **kwargs):
    """
    Solve a sparse linear system using the specified algorithm.
    """

    solver = JAX_ALGORITHM_DICT[algorithm]

    if not "jax" in algorithm:
        A = lx.MatrixLinearOperator(A)
    
    out = solver(A, b, **kwargs)

    if not "jax" in algorithm:
        out = out.value

    return out


# %% ../nbs/00_core.ipynb 13
def calculate_force(psi):
    h = 1 / (psi.shape[0] - 1)
    ny = psi.shape[1]
    
    return np.sum(
        2 * psi[1:-1, ny - 1] -
        5 * psi[1:-1, ny - 2] +
        4 * psi[1:-1, ny - 3] -
        psi[1:-1, ny - 4]
    ) / h


# %% ../nbs/00_core.ipynb 14
def calculate_velocity(psi):
    h = 1 / (psi.shape[0] - 1)

    nx = psi.shape[0]
    
    u = np.zeros_like(psi)
    v = np.zeros_like(psi)
    
    u[1:-1, 1:-1] = (psi[1:-1, 2:] - psi[1:-1, :-2]) / (2 * h)
    v[1:-1, 1:-1] = -(psi[2:, 1:-1] - psi[:-2, 1:-1]) / (2 * h)
    
    u[1:-1, -1] = np.sin(np.pi * np.arange(1, nx - 1) * h) ** 2
    
    return u, v




# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/uom-project',
                'doc_host': 'https://AugustasMacijauskas.github.io',
                'git_url': 'https://github.com/AugustasMacijauskas/uom-project',
                'lib_path': 'uom_project'},
  'syms': { 'uom_project.core': { 'uom_project.core.calculate_force': ('core.html#calculate_force', 'uom_project/core.py'),
                                  'uom_project.core.calculate_velocity': ('core.html#calculate_velocity', 'uom_project/core.py'),
                                  'uom_project.core.lu_solve': ('core.html#lu_solve', 'uom_project/core.py'),
                                  'uom_project.core.setup_poisson_problem': ('core.html#setup_poisson_problem', 'uom_project/core.py'),
                                  'uom_project.core.solve_sparse_linear_system': ( 'core.html#solve_sparse_linear_system',
                                                                                   'uom_project/core.py'),
                                  'uom_project.core.solve_sparse_linear_system_jax': ( 'core.html#solve_sparse_linear_system_jax',
                                                                                       'uom_project/core.py'),
                                  'uom_project.core.solve_sparse_linear_system_pytorch': ( 'core.html#solve_sparse_linear_system_pytorch',
                                                                                           'uom_project/core.py')},
            'uom_project.poisson_solvers': { 'uom_project.poisson_solvers.construct_laplacian_kernel_matrix': ( 'poisson_solvers.html#construct_laplacian_kernel_matrix',
                                                                                                                'uom_project/poisson_solvers.py'),
                                             'uom_project.poisson_solvers.f': ('poisson_solvers.html#f', 'uom_project/poisson_solvers.py'),
                                             'uom_project.poisson_solvers.get_jacobian': ( 'poisson_solvers.html#get_jacobian',
                                                                                           'uom_project/poisson_solvers.py'),
                                             'uom_project.poisson_solvers.newton_iterator': ( 'poisson_solvers.html#newton_iterator',
                                                                                              'uom_project/poisson_solvers.py'),
                                             'uom_project.poisson_solvers.poisson_gauss_seidel_with_sor_solver': ( 'poisson_solvers.html#poisson_gauss_seidel_with_sor_solver',
                                                                                                                   'uom_project/poisson_solvers.py'),
                                             'uom_project.poisson_solvers.poisson_newton_alternative_solver': ( 'poisson_solvers.html#poisson_newton_alternative_solver',
                                                                                                                'uom_project/poisson_solvers.py'),
                                             'uom_project.poisson_solvers.poisson_newton_solver': ( 'poisson_solvers.html#poisson_newton_solver',
                                                                                                    'uom_project/poisson_solvers.py'),
                                             'uom_project.poisson_solvers.poisson_non_iterative_solver': ( 'poisson_solvers.html#poisson_non_iterative_solver',
                                                                                                           'uom_project/poisson_solvers.py')},
            'uom_project.streamfunction_vorticity': { 'uom_project.streamfunction_vorticity.get_dw_dt': ( '02_streamfunction_vorticity_iterative_solver/streamfunction_vorticity_iterative_solver.html#get_dw_dt',
                                                                                                          'uom_project/streamfunction_vorticity.py'),
                                                      'uom_project.streamfunction_vorticity.streamfunction_vorticity_iterative_solver': ( '02_streamfunction_vorticity_iterative_solver/streamfunction_vorticity_iterative_solver.html#streamfunction_vorticity_iterative_solver',
                                                                                                                                          'uom_project/streamfunction_vorticity.py'),
                                                      'uom_project.streamfunction_vorticity.update_vorticity_bcs': ( '02_streamfunction_vorticity_iterative_solver/streamfunction_vorticity_iterative_solver.html#update_vorticity_bcs',
                                                                                                                     'uom_project/streamfunction_vorticity.py')},
            'uom_project.streamfunction_vorticity_jax': { 'uom_project.streamfunction_vorticity_jax.f_jax': ( 'streamfunction_vorticity_jax.html#f_jax',
                                                                                                              'uom_project/streamfunction_vorticity_jax.py'),
                                                          'uom_project.streamfunction_vorticity_jax.newton_iteration_step': ( 'streamfunction_vorticity_jax.html#newton_iteration_step',
                                                                                                                              'uom_project/streamfunction_vorticity_jax.py'),
                                                          'uom_project.streamfunction_vorticity_jax.newton_iterator_jax': ( 'streamfunction_vorticity_jax.html#newton_iterator_jax',
                                                                                                                            'uom_project/streamfunction_vorticity_jax.py'),
                                                          'uom_project.streamfunction_vorticity_jax.newton_solver_jax': ( 'streamfunction_vorticity_jax.html#newton_solver_jax',
                                                                                                                          'uom_project/streamfunction_vorticity_jax.py'),
                                                          'uom_project.streamfunction_vorticity_jax.reconstruct_w_jax': ( 'streamfunction_vorticity_jax.html#reconstruct_w_jax',
                                                                                                                          'uom_project/streamfunction_vorticity_jax.py')},
            'uom_project.streamfunction_vorticity_newton': { 'uom_project.streamfunction_vorticity_newton.f': ( 'streamfunction_vorticity_newton.html#f',
                                                                                                                'uom_project/streamfunction_vorticity_newton.py'),
                                                             'uom_project.streamfunction_vorticity_newton.get_jacobian': ( 'streamfunction_vorticity_newton.html#get_jacobian',
                                                                                                                           'uom_project/streamfunction_vorticity_newton.py'),
                                                             'uom_project.streamfunction_vorticity_newton.get_standard_basis_vector': ( 'streamfunction_vorticity_newton.html#get_standard_basis_vector',
                                                                                                                                        'uom_project/streamfunction_vorticity_newton.py'),
                                                             'uom_project.streamfunction_vorticity_newton.newton_iterator': ( 'streamfunction_vorticity_newton.html#newton_iterator',
                                                                                                                              'uom_project/streamfunction_vorticity_newton.py'),
                                                             'uom_project.streamfunction_vorticity_newton.newton_solver': ( 'streamfunction_vorticity_newton.html#newton_solver',
                                                                                                                            'uom_project/streamfunction_vorticity_newton.py'),
                                                             'uom_project.streamfunction_vorticity_newton.reconstruct_w': ( 'streamfunction_vorticity_newton.html#reconstruct_w',
                                                                                                                            'uom_project/streamfunction_vorticity_newton.py')},
            'uom_project.streamfunction_vorticity_pytorch': { 'uom_project.streamfunction_vorticity_pytorch.f_pytorch': ( 'streamfunction_vorticity_pytorch.html#f_pytorch',
                                                                                                                          'uom_project/streamfunction_vorticity_pytorch.py'),
                                                              'uom_project.streamfunction_vorticity_pytorch.newton_iterator_pytorch': ( 'streamfunction_vorticity_pytorch.html#newton_iterator_pytorch',
                                                                                                                                        'uom_project/streamfunction_vorticity_pytorch.py'),
                                                              'uom_project.streamfunction_vorticity_pytorch.newton_solver_pytorch': ( 'streamfunction_vorticity_pytorch.html#newton_solver_pytorch',
                                                                                                                                      'uom_project/streamfunction_vorticity_pytorch.py'),
                                                              'uom_project.streamfunction_vorticity_pytorch.reconstruct_w_pytorch': ( 'streamfunction_vorticity_pytorch.html#reconstruct_w_pytorch',
                                                                                                                                      'uom_project/streamfunction_vorticity_pytorch.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_streamfunction_vorticity_newton.ipynb.

# %% auto 0
__all__ = ['get_standard_basis_vector', 'get_jacobian', 'f', 'reconstruct_w', 'newton_iterator', 'newton_solver']

# %% ../nbs/03_streamfunction_vorticity_newton.ipynb 6
from . import core, poisson_solvers

import numpy as np
import scipy
from scipy import sparse

# %% ../nbs/03_streamfunction_vorticity_newton.ipynb 9
# Given the vorticity, solve the Poisson eqn. to find the streamfunction
def get_standard_basis_vector(size, i):
    vec = np.zeros(size)
    vec[i] = 1.0
    
    return vec


def get_jacobian(f, x, Re):
    N = int(np.sqrt(x.shape[0] // 2 + 1))
    h = 1 / N

    f_evaluated = f(x=x, Re=Re)
    
    # Jacobian is sparse
    return sparse.csr_matrix(
        np.vstack([(
            f(x=x + h*get_standard_basis_vector(size=x.shape[0], i=i), Re=Re) -
            f_evaluated
        ) for i in range(x.shape[0])])
    ).T


def f(x, Re, U_wall_top):
    N = int(np.sqrt(x.shape[0] // 2 + 1))
    h = 1 / N

    psi = x[:(N-1)**2].reshape(N-1, N-1)
    w_left   = x[(N-1)**2 + 0*(N-1) : (N-1)**2 + 1*(N-1)]
    w_right  = x[(N-1)**2 + 1*(N-1) : (N-1)**2 + 2*(N-1)]
    w_bottom = x[(N-1)**2 + 2*(N-1) : (N-1)**2 + 3*(N-1)]
    w_top    = x[(N-1)**2 + 3*(N-1) : (N-1)**2 + 4*(N-1)]
    w_middle = x[(N-1)**2 + 4*(N-1) :].reshape(N-1, N-1)


    # Calculate the equations coming from the Poisson equation
    f_poisson = -4 * psi + h ** 2 * w_middle
    f_poisson[1:, :] += psi[:-1, :]
    f_poisson[:-1, :] += psi[1:, :]
    f_poisson[:, :-1] += psi[:, 1:]
    f_poisson[:, 1:] += psi[:, :-1]


    # Calculate contributions coming from the vorticity transport equation
    
    # Calculate the sides first
    # y = 0, U_wall = 0
    f_w_bottom = h ** 2 * (w_middle[:, 0] + 3 * w_bottom) + 8 * psi[:, 0]
    # y = 1, U_wall is known here
    f_w_top = h ** 2 * (w_middle[:, -1] + 3 * w_top) + 8 * (
        h * U_wall_top + psi[:, -1]
    )
    # x = 0
    f_w_left = h ** 2 * (w_middle[0, :] + 3 * w_left) + 8 * psi[0, :]
    # x = 1
    f_w_right = h ** 2 * (w_middle[-1, :] + 3 * w_right) + 8 * psi[-1, :]

    f_w_middle = -4 * w_middle
    f_w_middle[:-1, :] += w_middle[1:, :]
    f_w_middle[-1:, :] += w_right
    f_w_middle[1:, :] += w_middle[:-1, :]
    f_w_middle[:1, :] += w_left
    f_w_middle[:, :-1] += w_middle[:, 1:]
    f_w_middle[:, -1] += w_top
    f_w_middle[:, 1:] += w_middle[:, :-1]
    f_w_middle[:, 0] += w_bottom

    f_w_middle[1:-1, 1:-1] += Re * (
        (psi[2:, 1:-1] - psi[:-2, 1:-1]) * (w_middle[1:-1, 2:] - w_middle[1:-1, :-2]) -
        (psi[1:-1, 2:] - psi[1:-1, :-2]) * (w_middle[2:, 1:-1] - w_middle[:-2, 1:-1])
    ) / 4
    f_w_middle[:1, 1:-1] += Re * (
        psi[1, 1:-1] * (w_middle[0, 2:] - w_middle[0, :-2]) -
        (psi[0, 2:] - psi[0, :-2]) * (w_middle[1, 1:-1] - w_left[1:-1])
    ) / 4
    f_w_middle[-1:, 1:-1] -= Re * (
        psi[-2, 1:-1] * (w_middle[-1, 2:] - w_middle[-1, :-2]) +
        (psi[-1, 2:] - psi[-1, :-2]) * (w_right[1:-1] - w_middle[-2, 1:-1])
    ) / 4
    f_w_middle[1:-1, 0] += Re * (
        (psi[2:, 0] - psi[:-2, 0]) * (w_middle[1:-1, 1] - w_bottom[1:-1]) -
        psi[1:-1, 1] * (w_middle[2:, 0] - w_middle[:-2, 0])
    ) / 4
    f_w_middle[1:-1, -1] += Re * (
        (psi[2:, -1] - psi[:-2, -1]) * (w_top[1:-1] - w_middle[1:-1, -2]) +
        psi[1:-1, -2] * (w_middle[2:, -1] - w_middle[:-2, -1])
    ) / 4
    f_w_middle[0, 0] += Re * (
        psi[1, 0] * (w_middle[0, 1] - w_bottom[0]) -
        psi[0, 1] * (w_middle[1, 0] - w_left[0])
    ) / 4
    f_w_middle[-1, 0] -= Re * (
        psi[-2, 0] * (w_middle[-1, 1] - w_bottom[-1]) +
        psi[-1, 1] * (w_right[0] - w_middle[-2, 0])
    ) / 4
    f_w_middle[0, -1] += Re * (
        psi[1, -1] * (w_top[0] - w_middle[0, -2]) +
        psi[0, -2] * (w_middle[1, -1] - w_left[-1])
    ) / 4
    f_w_middle[-1, -1] -= Re * (
        psi[-2, -1] * (w_top[-1] - w_middle[-1, -2]) -
        psi[-1, -2] * (w_right[-1] - w_middle[-2, -1])
    ) / 4

    return np.concatenate([
        f_poisson.flatten(), f_w_left, f_w_right, f_w_bottom, f_w_top, f_w_middle.flatten()
    ], axis=0)


def reconstruct_w(w_tmp, N):
    w = np.zeros((N+1, N+1))
    
    w[0, 1:-1] = w_tmp[0*(N-1):1*(N-1)]
    w[-1, 1:-1] = w_tmp[1*(N-1):2*(N-1)]
    w[1:-1, 0] = w_tmp[2*(N-1):3*(N-1)]
    w[1:-1, -1] = w_tmp[3*(N-1):4*(N-1)]
    w[1:-1, 1:-1] = w_tmp[4*(N-1):].reshape((N - 1, N - 1))
    
    return w


def newton_iterator(
    f, get_jacobian, N, Re,
    algorithm="base", TOL=1e-8, max_iter=10, quiet=True
):
    '''
        - f: evaluates the function given x, Re
        - get_jacobian: evaluates the Jacobian given N, h
        - N: number of grid points
        - h: grid size
        - Re: Reynolds number
    '''

    h = 1 / N

    n_iter = 0 # number of iterations

    # Initialization
    # Size (N - 1) ** 2         + (N + 1) ** 2    - 4
    # Size (for streamfunction) + (for vorticity) - (corners of vorticity)
    x = np.zeros((N - 1) ** 2 + (N + 1) ** 2 - 4)
    f_current = f(x=x, Re=Re)
    
    # Check if the initial guess is a solution
    f_norm = scipy.linalg.norm(f_current)
    if f_norm <= TOL:
        if not quiet:
            print(f"n_iter={n_iter}")

        return x, n_iter
    
    while n_iter < max_iter:
        n_iter += 1
        jacobian = get_jacobian(f=f, x=x, Re=Re)

        kwargs = {}
        if algorithm == "bicgstab": kwargs["tol"] = 1e-6
        dx = core.solve_sparse_linear_system(
            A=jacobian, b=-h * f_current, algorithm=algorithm, **kwargs
        )
        x_next = x + dx
        
        f_current = f(x=x_next, Re=Re)
        
        f_norm = scipy.linalg.norm(f_current)
        if not quiet:
            print(f"iter={n_iter}; residual={f_norm}; dx={scipy.linalg.norm(dx)}")
        if f_norm <= TOL:
            break
        
        x = x_next
        
    if not quiet:
        print(f"n_iter={n_iter}")
    
    return x_next, n_iter


# %% ../nbs/03_streamfunction_vorticity_newton.ipynb 10
def newton_solver(
    f, get_jacobian, N, Re,
    algorithm="base", TOL=1e-8, max_iter=10, quiet=True
):
    assert algorithm != "cg", (
        "Linear system is not positive definite, cg algorithm cannot be used"
    )

    solution, n_iter = newton_iterator(
        f=f, get_jacobian=get_jacobian, N=N, Re=Re,
        algorithm=algorithm,
        TOL=TOL, max_iter=max_iter, quiet=quiet
    )
    
    psi, w = solution[:(N - 1) ** 2], solution[(N - 1) ** 2:]
    
    # Get final psi
    psi = psi.reshape(N - 1, N - 1)
    psi = np.pad(psi, (1, 1), mode="constant", constant_values=0)
    
    # Get final w
    w = reconstruct_w(w_tmp=w, N=N)
    w = w.reshape(N + 1, N + 1)
    
    return w, psi, n_iter




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_streamfunction_vorticity_iterative_solver/02_streamfunction_vorticity_iterative_solver.ipynb.

# %% auto 0
__all__ = ['streamfunction_vorticity_iterative_solver']

# %% ../nbs/02_streamfunction_vorticity_iterative_solver/02_streamfunction_vorticity_iterative_solver.ipynb 6
from . import poisson_solvers

import numpy as np

# %% ../nbs/02_streamfunction_vorticity_iterative_solver/02_streamfunction_vorticity_iterative_solver.ipynb 9
def update_vorticity_bcs(w, psi, U_wall_top, nx, ny, h):
    # First calculate BCs to 1st order

    # y = 0: U_wall = 0
    w[1:nx - 1, 0] = -(
        (psi[1:nx - 1, 1] - psi[1:nx - 1, 0]) / h - 0
    ) / (0.5 * h)
    # y = 1: U_wall_top is given here
    w[1:nx - 1, ny - 1] = -(
        U_wall_top -
        (psi[1:nx - 1, ny - 1] - psi[1:nx - 1, ny - 2]) / h
    ) / (0.5 * h)
    
    # x = 0: U_wall = 0
    w[0, 1:ny - 1] = (
        -(psi[1, 1:ny - 1] - psi[0, 1:ny - 1]) / h - 0
    ) / (0.5 * h)
    # x = 1: U_wall = 0
    w[nx - 1, 1:ny - 1] = (
        0 - (-(psi[nx - 1, 1:ny - 1] - psi[nx - 2, 1:ny - 1]) / h)
    ) / (0.5 * h)

    
    # Compute the 2nd order correction to the BCs
    w[1:nx - 1, 0] = (4 * w[1:nx - 1, 0] - w[1:nx - 1, 1]) / 3 # y = 0
    w[1:nx - 1, ny - 1] = (
        4 * w[1:nx - 1, ny - 1] - w[1:nx - 1, ny - 2]
    ) / 3 # y = 1
    
    w[0, 1:ny - 1] = (4 * w[0, 1:ny - 1] - w[1, 1:ny - 1]) / 3 # x = 0
    w[nx - 1, 1:ny - 1] = (
        4 * w[nx - 1, 1:ny - 1] - w[nx - 2, 1:ny - 1]
    ) / 3 # x = 1

    return w


def get_dw_dt(dw_dt, w, psi, Re, nx, ny, h):
    dw_dt[1:nx - 1, 1:ny - 1] = (
        -(
            (psi[1:nx - 1, 2:ny] - psi[1:nx - 1, 0:ny - 2]) *
            (w[2:nx, 1:ny - 1] - w[0:nx - 2, 1:ny - 1])
        ) / (4 * h ** 2) + 
        (
            (psi[2:nx, 1:ny - 1] - psi[0:nx - 2, 1:ny - 1]) *
            (w[1:nx - 1, 2:ny] - w[1:nx - 1, 0:ny - 2])
        ) / (4 * h ** 2) + 
        (
            w[2:nx, 1:ny - 1] + w[0:nx - 2, 1:ny - 1] +
            w[1:nx - 1, 2:ny] + w[1:nx - 1, 0:ny - 2] -
            4 * w[1:nx - 1 , 1:ny - 1]
        ) / (Re * h ** 2)
    )

    return dw_dt


# %% ../nbs/02_streamfunction_vorticity_iterative_solver/02_streamfunction_vorticity_iterative_solver.ipynb 10
# Incorporates both the Poisson problem and time-stepping for the vorticity
def streamfunction_vorticity_iterative_solver(
    N, Re, tfinal, U_wall_top, dt=None,
    algorithm="base", print_every=0.0
):
    nx = ny = N + 1 # i.e. N = 10
    h = 1 / N # h = dx (=dy) = 1 / N; x-size is 1, y-size is ny/nx
    
    w = np.zeros(shape=(nx, ny)) # vorticity
    psi = np.zeros(shape=(nx, ny)) # streamfunction
    dw_dt = np.zeros(shape=(nx, ny)) # dw/dt    
    
    if dt is None:
        dt = np.round(0.2 * Re * h ** 2, 8) # 0.8 * marginal value
    
    omega_mid = []
    
    t = 0
    while t < tfinal:
        # Start by solving the Poisson problem
        psi = poisson_solvers.poisson_non_iterative_solver(
            w, algorithm=algorithm
        )

        # Now time-step vorticity
        w = update_vorticity_bcs(w, psi, U_wall_top, nx, ny, h)

        # We can now find dw/dt
        dw_dt = get_dw_dt(dw_dt, w, psi, Re, nx, ny, h)

        # Finally, update w
        w[1:nx-1, 1:ny-1] = w[1:nx-1, 1:ny-1] + dw_dt[1:nx-1, 1:ny-1] * dt
        
        # One pass done, increment time
        t = np.round(t + dt, 8)
        
        # Print
        if print_every > 0:
            if 0 <= np.round(t % print_every, 8) < dt:
                print(f"t={t:.5f}; w(0.5, 0.5)={w[nx // 2, ny // 2]}")
        
        omega_mid.append(w[nx // 2, ny // 2])
    
    return w, psi, np.array(omega_mid)

